@@ -1,27 +1,13 @@
-/*=============================================================================
-
-    This file is part of ARB.
-
-    ARB is free software; you can redistribute it and/or modify
-    it under the terms of the GNU General Public License as published by
-    the Free Software Foundation; either version 2 of the License, or
-    (at your option) any later version.
-
-    ARB is distributed in the hope that it will be useful,
-    but WITHOUT ANY WARRANTY; without even the implied warranty of
-    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
-    GNU General Public License for more details.
-
-    You should have received a copy of the GNU General Public License
-    along with ARB; if not, write to the Free Software
-    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
-
-=============================================================================*/
-/******************************************************************************
-
+/*
     Copyright (C) 2014 Fredrik Johansson
 
-******************************************************************************/
+    This file is part of Arb.
+
+    Arb is free software: you can redistribute it and/or modify it under
+    the terms of the GNU Lesser General Public License (LGPL) as published
+    by the Free Software Foundation; either version 2.1 of the License, or
+    (at your option) any later version.  See <http://www.gnu.org/licenses/>.
+*/
 
 #ifndef MAG_H
 #define MAG_H
@@ -32,11 +18,18 @@
 #define MAG_INLINE static __inline__
 #endif
 
+#include <stdio.h>
 #include <math.h>
 #include "flint/flint.h"
 #include "flint/fmpz.h"
 #include "fmpz_extras.h"
 
+#ifndef flint_abort
+#if __FLINT_RELEASE <= 20502
+#define flint_abort abort
+#endif
+#endif
+
 #ifdef __cplusplus
 extern "C" {
 #endif
@@ -152,7 +145,7 @@ __mag_fixmul32(mp_limb_t x, mp_limb_t y)
     if (MAG_MAN(rr) != 0 && FLINT_BIT_COUNT(MAG_MAN(rr)) != MAG_BITS) \
     { \
         flint_printf("FAIL: wrong number of bits in mantissa!\n"); \
-        abort(); \
+        flint_abort(); \
     }
 
 /* Note: assumes mantissa either has the right number of bits, or
@@ -163,7 +156,7 @@ __mag_fixmul32(mp_limb_t x, mp_limb_t y)
 #define MAG_ADJUST_ONE_TOO_LARGE(x) \
     do { \
         mp_limb_t __t = MAG_MAN(x) >> MAG_BITS; \
-        MAG_MAN(x) = (MAG_MAN(x) >> __t) + __t; \
+        MAG_MAN(x) = (MAG_MAN(x) >> __t) + (__t & MAG_MAN(x)); \
         if (__t) \
             fmpz_add_ui(MAG_EXPREF(x), MAG_EXPREF(x), __t); \
     } while (0)
@@ -171,7 +164,7 @@ __mag_fixmul32(mp_limb_t x, mp_limb_t y)
 #define MAG_FAST_ADJUST_ONE_TOO_LARGE(x) \
     do { \
         mp_limb_t __t = MAG_MAN(x) >> MAG_BITS; \
-        MAG_MAN(x) = (MAG_MAN(x) >> __t) + __t; \
+        MAG_MAN(x) = (MAG_MAN(x) >> __t) + (__t & MAG_MAN(x)); \
         MAG_EXP(x) += __t; \
     } while (0)
 
@@ -216,11 +209,7 @@ mag_init_set(mag_t x, const mag_t y)
     MAG_MAN(x) = MAG_MAN(y);
 }
 
-MAG_INLINE void
-mag_clear(mag_t x)
-{
-    fmpz_clear(MAG_EXPREF(x));
-}
+void mag_clear(mag_t x);
 
 MAG_INLINE void
 mag_swap(mag_t x, mag_t y)
@@ -304,6 +293,10 @@ void mag_add(mag_t z, const mag_t x, const mag_t y);
 
 void mag_add_lower(mag_t z, const mag_t x, const mag_t y);
 
+void mag_add_ui(mag_t z, const mag_t x, ulong y);
+
+void mag_add_ui_2exp_si(mag_t z, const mag_t x, ulong y, slong e);
+
 void mag_div(mag_t z, const mag_t x, const mag_t y);
 
 MAG_INLINE void
@@ -337,6 +330,8 @@ mag_mul_2exp_fmpz(mag_t z, const mag_t x, const fmpz_t y)
     }
 }
 
+void mag_sub(mag_t z, const mag_t x, const mag_t y);
+
 void mag_sub_lower(mag_t z, const mag_t x, const mag_t y);
 
 /* Fast versions (no infs/nans, small exponents). Note that this
@@ -473,7 +468,41 @@ mag_fast_add_2exp_si(mag_t z, const mag_t x, slong e)
     }
 }
 
+/* requires that x is positive and finite */
+#define MAG_SET_D_2EXP(man, exp, x, xexp) \
+    do { \
+        int __cexp; \
+        double __x; \
+        int __fix; \
+        mp_limb_t __man; \
+        __x = frexp((x), &__cexp); \
+        __man = (mp_limb_t)(__x * (double)(LIMB_ONE << MAG_BITS)) + 1; \
+        __fix = __man >> (MAG_BITS); \
+        __man = (__man >> __fix) + __fix; \
+        (man) = __man; \
+        (exp) = (xexp) + __cexp + __fix; \
+    } while (0);
+
+/* requires that x is positive and finite */
+#define MAG_SET_D_2EXP_LOWER(man, exp, x, xexp) \
+    do { \
+        int __cexp; \
+        double __x; \
+        int __fix; \
+        mp_limb_t __man; \
+        __x = frexp((x), &__cexp); \
+        __man = (mp_limb_t)(__x * (double)(LIMB_ONE << MAG_BITS)) - 1; \
+        __fix = __man < MAG_ONE_HALF; \
+        __man = (__man << __fix); \
+        (man) = __man; \
+        (exp) = (xexp) + __cexp - __fix; \
+    } while (0);
+
+void mag_set_d(mag_t z, double x);
+void mag_set_d_lower(mag_t z, double x);
 void mag_set_d_2exp_fmpz(mag_t z, double c, const fmpz_t exp);
+void mag_set_d_2exp_fmpz_lower(mag_t z, double c, const fmpz_t exp);
+
 void mag_set_fmpz_2exp_fmpz(mag_t z, const fmpz_t man, const fmpz_t exp);
 
 #include "fmpr.h"
@@ -486,9 +515,21 @@ void mag_randtest_special(mag_t x, flint_rand_t state, slong expbits);
 
 void mag_randtest(mag_t x, flint_rand_t state, slong expbits);
 
-void mag_print(const mag_t x);
+void mag_fprint(FILE * file, const mag_t x);
 
-void mag_printd(const mag_t x, slong d);
+void mag_fprintd(FILE * file, const mag_t x, slong d);
+
+MAG_INLINE void
+mag_print(const mag_t x)
+{
+    mag_fprint(stdout, x);
+}
+
+MAG_INLINE void
+mag_printd(const mag_t x, slong d)
+{
+    mag_fprintd(stdout, x, d);
+}
 
 void mag_get_fmpq(fmpq_t y, const mag_t x);
 
@@ -535,17 +576,10 @@ _mag_vec_clear(mag_ptr v, slong n)
     flint_free(v);
 }
 
-MAG_INLINE void mag_set_d(mag_t z, double x)
-{
-    fmpz_t e;
-    fmpz_init(e);
-    mag_set_d_2exp_fmpz(z, x, e);
-    fmpz_clear(e);
-}
-
-/* TODO: test/document */
 double mag_get_d(const mag_t z);
 
+double mag_get_d_log2_approx(const mag_t x);
+
 /* TODO: document */
 double mag_d_log_upper_bound(double x);
 double mag_d_log_lower_bound(double x);
@@ -565,12 +599,15 @@ mag_exp(mag_t y, const mag_t x)
 void mag_expm1(mag_t y, const mag_t x);
 void mag_exp_tail(mag_t z, const mag_t x, ulong N);
 
+void mag_expinv(mag_t res, const mag_t x);
+
 void mag_pow_ui(mag_t z, const mag_t x, ulong e);
 void mag_pow_ui_lower(mag_t z, const mag_t x, ulong e);
 void mag_pow_fmpz(mag_t z, const mag_t x, const fmpz_t e);
 
 void mag_fac_ui(mag_t z, ulong n);
 void mag_rfac_ui(mag_t z, ulong n);
+void mag_bin_uiui(mag_t res, ulong n, ulong k);
 
 /* TODO: test */
 void mag_bernoulli_div_fac_ui(mag_t z, ulong n);
@@ -579,6 +616,7 @@ void mag_bernoulli_div_fac_ui(mag_t z, ulong n);
 void mag_set_fmpz_2exp_fmpz_lower(mag_t z, const fmpz_t man, const fmpz_t exp);
 
 void mag_sqrt(mag_t y, const mag_t x);
+void mag_sqrt_lower(mag_t y, const mag_t x);
 void mag_rsqrt(mag_t y, const mag_t x);
 
 void mag_root(mag_t y, const mag_t x, ulong n);
@@ -591,6 +629,8 @@ void mag_polylog_tail(mag_t u, const mag_t z, slong sigma, ulong d, ulong N);
 
 void mag_geom_series(mag_t res, const mag_t x, ulong n);
 
+void mag_hurwitz_zeta_uiui(mag_t res, ulong s, ulong a);
+
 void mag_set_ui(mag_t z, ulong x);
 void mag_set_ui_lower(mag_t z, ulong x);
 
@@ -678,6 +718,16 @@ mag_div_fmpz(mag_t z, const mag_t x, const fmpz_t y)
     mag_clear(t);
 }
 
+MAG_INLINE void mag_const_pi(mag_t res)
+{
+    mag_set_ui_2exp_si(res, 843314857, -28);
+}
+
+MAG_INLINE slong mag_allocated_bytes(const mag_t x)
+{
+    return fmpz_allocated_bytes(MAG_EXPREF(x));
+}
+
 #ifdef __cplusplus
 }
 #endif
